import logging
from logging import getLogger
from typing import List
from pathlib import Path

from .config import CONFIG as _CONFIG
from .util import create_temporary_docker_volume, detect_free_gpu
from .volume import copy_input_to, copy_output_from
from .preprocessing import apply_super_resolution
from .ocr import run_tesseract, run_pero_ocr, combine_results

import click
import docker


LOGGER = getLogger(__name__)
CONFIG = _CONFIG['cli']


@click.command()
@click.argument('input-dir',
                type=click.Path(exists=True, dir_okay=True, file_okay=False),
                required=True)
@click.argument('output-dir',
                type=click.Path(exists=None, dir_okay=True, file_okay=False),
                required=True)
@click.option('--pero-ocr/--no-pero-ocr',
              help='Use PERO in addition to Tesseract for OCR (recommended, requires GPU)',
              default=bool(CONFIG.getint('pero_ocr')),
              is_flag=True,
              required=False)
@click.option('--super-resolution/--no-super-resolution',
              help='Use image super-resolution before the OCR (recommended, requires GPU)',
              default=bool(CONFIG.getint('super_resolution')),
              is_flag=True,
              required=False)
@click.option('--gpus',
              help=('Either comma-separated PCI BUS IDs of NVIDIA GPUs that will be used for image super-resolution '
                    'and PERO OCR, `auto` for an automatic selection of a free GPU, or `all` for the use of all GPUs'),
              default=CONFIG['gpus'],
              required=False)
def main(input_dir: str, output_dir: str, pero_ocr: bool, super_resolution: bool, gpus: str) -> None:

    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')

    client = docker.from_env()

    def check_docker_image(name: str) -> None:
        if not client.images.list(name):
            raise ValueError(f'Docker image {name} not found. Install it with make build.')

    check_docker_image('ahisto/empty')
    check_docker_image('ahisto/ocr-eval')
    check_docker_image('ahisto/ocr-fileformat')
    check_docker_image('ahisto/tesseract')
    check_docker_image('ahisto/waifu2x')
    check_docker_image('ahisto/pero-ocr')

    input_dir = Path(input_dir)
    input_images = sorted(
        filename.relative_to(input_dir)
        for filename
        in input_dir.glob('**/*.*')
    )

    output_dir = Path(output_dir)

    if gpus == 'auto':
        gpus = detect_free_gpu()

    run_ocr(input_dir, input_images, output_dir, pero_ocr, super_resolution, gpus)


def run_ocr(input_dir: Path, input_images: List[Path], output_dir: Path,
            pero_ocr: bool, super_resolution: bool, gpus: str) -> None:
    client = docker.from_env()
    with create_temporary_docker_volume(client) as input_volume, \
            create_temporary_docker_volume(client) as postprocessing_volume, \
            create_temporary_docker_volume(client) as tesseract_volume, \
            create_temporary_docker_volume(client) as pero_ocr_volume, \
            create_temporary_docker_volume(client) as output_volume:

        copy_input_to(client, input_volume, input_dir, input_images)

        if super_resolution:
            apply_super_resolution(client, input_volume, postprocessing_volume, gpus)
        else:
            LOGGER.info('Skipping image super-resolution')
            postprocessing_volume = input_volume

        run_tesseract(client, postprocessing_volume, tesseract_volume)

        if pero_ocr:
            run_pero_ocr(client, input_volume, pero_ocr_volume, gpus)
            combine_results(client, tesseract_volume, pero_ocr_volume, output_volume)
        else:
            LOGGER.info('Skipping PERO OCR')
            output_volume = tesseract_volume

        copy_output_from(client, output_volume, input_images, output_dir)
